import io
import base64
from PIL import ImageDraw, Image

def cropbox(xmin, ymin, xmax, ymax, img_size, ratio=1.5, make_square=False):
    if xmin < 0 or ymin <0 or xmax < 0 or ymax < 0:
        raise Exception
    w, h = img_size
    if xmin > w or ymin > h or xmax > w or ymax > h:
        raise Exception

    xc = xmin + (xmax - xmin) / 2
    yc = ymin + (ymax - ymin) / 2
    w = xmax - xmin
    h = ymax - ymin
    nw = w * ratio
    nh = h * ratio

    if make_square:
        if nw > nh:
            nh = nw
        else:
            nw = nh

    nxmin = max(xc - (nw / 2), 0)
    nymin = max(yc - (nh / 2), 0)

    nxmax = min(xc + (nw / 2), img_size[0])
    nymax = min(yc + (nh / 2), img_size[1])

    return nxmin, nymin, nxmax, nymax


def image_to_base64(img):
    output_buffer = io.BytesIO()
    img.save(output_buffer, format='JPEG')
    byte_data = output_buffer.getvalue()
    base64_str = base64.b64encode(byte_data)
    return base64_str

def base64_to_image(base64_str):
    return Image.open(io.BytesIO(base64.b64decode(base64_str)))

def draw_bounding_box_on_image(image, xmin, ymin, xmax, ymax,
                               color='red',
                               text='',
                               thickness=4):
    draw = ImageDraw.Draw(image)
    draw.rectangle([(xmin, ymin), (xmax, ymax)], outline=color, width=thickness)
    draw.text((xmin, ymin), text)
    return image



import math
from PIL import Image
import torch
from argparse import ArgumentParser
from torchvision import transforms as T
from .models.wrapper import CustomCLIPWrapper
import torch.nn.functional as F
from transformers import AutoTokenizer, AutoModel
import PIL
import soco_clip.clip as clip
import os
import cv2
import shutil
import random


class CLIP(object):
    def __init__(self, clip_model="ViT-B-16", model_path = None, tokenizer=None, txt_encoder=None):
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer) if tokenizer else None  #model_path = "zh_roberta_vit16.ckpt", tokenizer="hfl/chinese-roberta-wwm-ext", txt_encoder="hfl/chinese-roberta-wwm-ext"
        txt_encoder = AutoModel.from_pretrained(txt_encoder) if txt_encoder else None
        self.model = CustomCLIPWrapper.load_from_checkpoint(checkpoint_path=model_path, text_encoder=txt_encoder, minibatch_size=64, avg_word_embs=True, clip_model=clip_model ) if model_path else None
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.image_encoder,self.preprocess = clip.load(clip_model, device=self.device) 
        if self.model:
            self.model.eval()

    def texts(self, query):
        if self.tokenizer:
            tk = self.tokenizer(query, return_tensors='pt', padding=True, truncation=True)
            y_hat = self.model.change_text_emb(self.model.encode_text(tk))
            return F.normalize(y_hat.to(self.device))
        else:
            text_input = clip.tokenize(query).to(self.device)
            text_embedding = self.image_encoder.encode_text(text_input)
            return text_embedding

    def image(self, im):
        if type(im) == str:
            image = Image.open(im)
        else:
            image = im
        image_input = self.preprocess(image).unsqueeze(0).to(self.device)
        img = self.image_encoder.encode_image(image_input)
        return img.to(torch.float32)
    
    def images(self, urls):
        temp = []
        for url in urls:
            if type(url) == str:
                image = Image.open(url)
            else:
                image = url
            image_input = self.preprocess(image).unsqueeze(0).to(self.device)
            img = self.image_encoder.encode_image(image_input)
            temp.append(img[0])
        return torch.stack(temp, dim=0).to(torch.float32)

    def video(self,video_file, max_frames=50, rate=1000):
        pathOut = "tteemmpp_"+str(random.randint(0,1000000000))
        shutil.rmtree(pathOut,ignore_errors=True, onerror=None)
        os.makedirs(pathOut)
        count = 0
        vidcap = cv2.VideoCapture(video_file)
        success,image = vidcap.read()
        success = True
        paths = []
        while success:
            try:
                vidcap.set(cv2.CAP_PROP_POS_MSEC,(count*rate))    # added this line
                success,image = vidcap.read()
                if success:
                    path = os.path.join(pathOut, "frame%d.jpg" % count)     # save frame as JPEG file
                    cv2.imwrite( os.path.join(pathOut, "frame%d.jpg" % count), image)     # save frame as JPEG file
                    count = count + 1
                    paths.append(path)
            except Exception as e:
                print (e)
                continue
        if max_frames > len(paths):
            y = self.images(paths)
        else:
            y = self.images(random.choices(paths, k=max_frames))
        return y

    def video_texts(self,video_file, texts, max_frames=50, rate=1000):
        pathOut = "tteemmpp_"+str(random.randint(0,1000000000))
        shutil.rmtree(pathOut,ignore_errors=True, onerror=None)
        os.makedirs(pathOut)
        count = 0
        vidcap = cv2.VideoCapture(video_file)
        success,image = vidcap.read()
        success = True
        paths = []
        while success:
            try:
                vidcap.set(cv2.CAP_PROP_POS_MSEC,(count*rate))    # added this line
                success,image = vidcap.read()
                if success:
                    path = os.path.join(pathOut, "frame%d.jpg" % count)     # save frame as JPEG file
                    cv2.imwrite( os.path.join(pathOut, "frame%d.jpg" % count), image)     # save frame as JPEG file
                    count = count + 1
                    paths.append(path)
            except Exception as e:
                print (e)
                continue
        if max_frames > len(paths):
            y = self.images(paths)
        else:
            y = self.images(random.choices(paths, k=max_frames))
        t = self.texts(texts)
        z = t@y.mean(0, True).T.half()[:,0]
        maxindex = z.argmax()
        shutil.rmtree(pathOut,ignore_errors=True, onerror=None)
        return {"score":z.softmax(0)[maxindex].cpu().detach().item(), "pred":texts[maxindex]}




from soco_clip import CLIP


if __name__ == "__main__":
    import glob
    import numpy as np
    X = CLIP("ViT-L-14.pt")
    #kmodel = CLIP("ViT-B-16","en_roberta_vit16.ckpt","roberta-base","roberta-base")
    #kmodel = CLIP("ViT-B-16","zh_roberta_vit16.ckpt","hfl/chinese-roberta-wwm-ext","hfl/chinese-roberta-wwm-ext")
    #z = kmodel.text_encode(["冬天","冬天"])
    #i = X.image_encodes(["/mnt/lambda_data/KL/CUHK-PEDES/imgs/test_query/p20_s39.jpg","/mnt/lambda_data/KL/CUHK-PEDES/imgs/test_query/p53_s102.jpg"])
    #print (X.clip_text_encode(["i am a boy"]))
    #print (z@i.T)

    print (X.image("sample.png"))
    print (X.video("a.mp4"))
    print (X.video_texts("b.mp4",["play","washing","swimming"]))



import torch
import torch.nn as nn
import torch.nn.functional as F
import pytorch_lightning as pl
import numpy as np
import math
import yaml
import copy
from .model import CLIP

import soco_clip.clip as clip

class CLIPWrapper(pl.LightningModule):
    def __init__(self,
                 model_name: str,
                 config: dict,
                 minibatch_size: int
                 ):
        """A lightning wrapper for a CLIP model as specified in the paper.

        Args:
            model_name (str): A case sensitive visual model name.
            config (dict): A dictionary containing the CLIP instantiation parameters.
        """
        super().__init__()

        self.model_name = model_name
        self.model = CLIP(**config)
        self.minibatch_size = minibatch_size
        self.isViT = 'ViT' in self.model_name

        self.automatic_optimization = False

    def forward(self, image, text):
        return self.model(image, text)

    # Training loss: https://github.com/openai/CLIP/issues/83
    # Mini-batching thanks to https://github.com/crowsonkb / https://twitter.com/RiversHaveWings
    # Multi-GPU support: https://github.com/MicPie/clasp
    def training_step(self, train_batch, idx):
        # get optimizers and scheduler
        optimizer = self.optimizers()

        image, text = train_batch
        n = math.ceil(len(image) // self.minibatch_size)
        image_mbs = torch.chunk(image, n)
        text_mbs = torch.chunk(text, n)

        # calculate original statistics
        with torch.no_grad():
            ims = [F.normalize(self.model.encode_image(im), dim=1) for im in image_mbs]
            txt = [F.normalize(self.model.encode_text(t), dim=1) for t in text_mbs]
            # gather from all GPUs
            ims = self.all_gather(torch.cat(ims))
            txt = self.all_gather(torch.cat(txt))

            if len(ims.shape) == 3:
                ims = list(ims)
                txt = list(txt)
            else:
                ims = [ims]
                txt = [txt]

            image_logits = torch.cat(ims) @ torch.cat(txt).t() * self.model.logit_scale.exp()
            ground_truth = torch.arange(len(image_logits)).type_as(image_logits).long()
            loss = (F.cross_entropy(image_logits, ground_truth) + F.cross_entropy(image_logits.t(), ground_truth)).div(2)
            acc_i = (torch.argmax(image_logits, 1) == ground_truth).sum()
            acc_t = (torch.argmax(image_logits, 0) == ground_truth).sum()
            self.log_dict({'loss': loss / len(ims), 'acc': (acc_i + acc_t) / 2 / len(image) / len(ims)}, prog_bar=True)

        if isinstance(optimizer, list):
            optimizer = optimizer[0]
        optimizer.zero_grad()

        # image loss
        for j, mb in enumerate(image_mbs):
            images_tmp = copy.deepcopy(ims)
            images_tmp[self.global_rank][j*self.minibatch_size:(j+1)*self.minibatch_size] = F.normalize(self.model.encode_image(mb), dim=1)
            image_logits = torch.cat(images_tmp) @ torch.cat(txt).t() * self.model.logit_scale.exp()
            ground_truth = torch.arange(len(image_logits)).type_as(image_logits).long()
            loss = (F.cross_entropy(image_logits, ground_truth) + F.cross_entropy(image_logits.t(), ground_truth))/2
            self.manual_backward(loss)

        # text loss
        for j, mb in enumerate(text_mbs):
            text_tmp = copy.deepcopy(txt)
            text_tmp[self.global_rank][j*self.minibatch_size:(j+1)*self.minibatch_size] = F.normalize(self.model.encode_text(mb), dim=1)
            image_logits = torch.cat(ims) @ torch.cat(text_tmp).t() * self.model.logit_scale.exp()
            loss = (F.cross_entropy(image_logits, ground_truth) + F.cross_entropy(image_logits.t(), ground_truth))/2
            self.manual_backward(loss)

        optimizer.step()
        lr_scheduler = self.lr_schedulers()
        lr_scheduler.step()
        self.model.logit_scale.data.clamp_(-np.log(100), np.log(100))

    def validation_step(self, val_batch, idx):
        image, text = val_batch
        image_logits, text_logits = self.forward(image, text)
        ground_truth = torch.arange(len(image_logits))
        loss = (F.cross_entropy(image_logits, ground_truth) + F.cross_entropy(text_logits, ground_truth)).div(2)
        self.log('val_loss', loss)

    def configure_optimizers(self):
        lr = {
            "RN50": 5e-4,
            "RN101": 5e-4,
            "RN50x4": 5e-4,
            "RN50x16": 4e-4,
            "RN50x64": 3.6e-4,
            "ViT-B/32": 5e-4,
            "ViT-B/16": 5e-4,
            "ViT-L/14": 4e-4,
            "ViT-L/14-336px": 2e-5
        }[self.model_name]

        optimizer = torch.optim.AdamW(
            self.model.parameters(),
            lr=lr,
            betas=(
                0.9,
                0.98 if self.isViT else 0.999
            ),
            eps=1e-6 if self.isViT else 1e-8,
            weight_decay=0.2
        )

        # TODO Watch: https://github.com/openai/CLIP/issues/107
        lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
            optimizer,
            T_0=2000
        )

        return {'optimizer': optimizer, 'lr_scheduler': lr_scheduler}


class CustomCLIPWrapper(CLIPWrapper):
    def __init__(self,
                 text_encoder,
                 minibatch_size,
                 learning_rate=3e-3,
                 kl_coeff=1.0,
                 avg_word_embs=False,
                 clip_model="ViT-B-16"
                 ):
        #with open('soco_clip/models/configs/RN.yaml') as fin:
        ##with open('soco_clip/models/configs/ViT.yaml') as fin:
        #    config = yaml.safe_load(fin)['RN50']
        #    print (config)
        #   config = yaml.safe_load(fin)['ViT-B/16']
        #config = {'embed_dim': 1024, 'image_resolution': 224, 'vision_layers': [3, 4, 6, 3], 'vision_width': 256, 'vision_patch_size': 0, 'context_length': 77, 'vocab_size': 49408, 'transformer_width': 512, 'transformer_heads': 8, 'transformer_layers': 12}
        config = {'embed_dim': 1024, 'image_resolution': 224, 'vision_layers': [3, 4, 6, 3], 'vision_width': 256, 'vision_patch_size': 0, 'context_length': 77, 'vocab_size': 49408, 'transformer_width': 512, 'transformer_heads': 8, 'transformer_layers': 12}
        self.input_resolution = 224
        self.context_length = 77
        self.vocab_size = 49408


        super().__init__('RN50', config, minibatch_size)
        #super().__init__('ViT-B/16', config, minibatch_size)
        del self.model.visual
        del self.model.transformer
        #self.model.visual = image_encoder
        device = "cuda" if torch.cuda.is_available() else "cpu"
        with torch.no_grad():
            self.image_encoder,self.preprocess = clip.load(clip_model, device=device)
        self.model.transformer = text_encoder
        self.learning_rate = learning_rate
        self.avg_word_embs = avg_word_embs
        self.sink_temp = nn.Parameter(torch.ones([]) * np.log(1 / 0.07))
        self.change_text_emb = nn.Linear(768,512)

        # init self-distillation model
        self.teacher = copy.deepcopy(self.model)
        self.kl_coeff = kl_coeff
        def convert_models_to_fp32(model): 
            for p in model.parameters(): 
                p.data = p.data.float() 
                p.grad.data = p.grad.data.float() 
        #clip.model.convert_weights(self.image_encoder)


    def training_step(self, train_batch, idx):
        # get optimizers and scheduler
        optimizer = self.optimizers()

        image, text = train_batch
        n = math.ceil(len(image) // self.minibatch_size)

        image_mbs = torch.chunk(image, n)

        text_mbs_ids = torch.chunk(torch.arange(len(image)), n)

        # adjust embedding dictionaries
        text_mbs = []
        for s in text_mbs_ids:
            d = {}
            for key in list(text.keys()):
                d[key] = text[key][s]
            text_mbs.append(d)

        # calculate original statistics
        with torch.no_grad():
            ims = [F.normalize(self.image_encoder.encode_image(im), dim=1) for im in image_mbs]
            txt = [F.normalize(self.change_text_emb(self.encode_text(t)), dim=1) for t in text_mbs]
            # gather from all GPUs
            ims = self.all_gather(torch.cat(ims))
            txt = self.all_gather(torch.cat(txt))

            if len(ims.shape) == 3:
                ims = list(ims)
                txt = list(txt)
            else:
                ims = [ims]
                txt = [txt]

            image_logits_notemp = torch.cat(ims) @ torch.cat(txt).t()
            image_logits = image_logits_notemp * self.model.logit_scale.exp()
            ground_truth = torch.arange(len(image_logits)).type_as(image_logits).long()
            loss = (F.cross_entropy(image_logits, ground_truth) + F.cross_entropy(image_logits.t(), ground_truth)).div(2)
            acc_i = (torch.argmax(image_logits, 1) == ground_truth).sum()
            acc_t = (torch.argmax(image_logits, 0) == ground_truth).sum()
            # calculate teacher
            teacher_ims = [F.normalize(self.image_encoder.encode_image(im), dim=1) for im in image_mbs]
            teacher_txt = [F.normalize(self.change_text_emb(self.encode_text(t, teacher=True)), dim=1) for t in text_mbs]

            teacher_ims = self.all_gather(torch.cat(teacher_ims))
            teacher_txt = self.all_gather(torch.cat(teacher_txt))

            if len(teacher_ims.shape) == 3:
                teacher_ims = list(teacher_ims)
                teacher_txt = list(teacher_txt)
            else:
                teacher_ims = [teacher_ims]
                teacher_txt = [teacher_txt]

            sim_ii, sim_tt, sim_it, sim_ti = self.compute_similarities(torch.cat(teacher_ims), torch.cat(teacher_txt))

            # optimal transport
            img_cost = - (sim_ii + sim_tt + sim_it)
            txt_cost = - (sim_ii + sim_tt + sim_ti)
            img_target = self.sinkhorn(img_cost)
            txt_target = self.sinkhorn(txt_cost)
            loss += (F.kl_div(image_logits_notemp * self.sink_temp, img_target) + F.kl_div(image_logits_notemp.t() * self.sink_temp, txt_target)) / 2 * self.kl_coeff
            self.log_dict({'loss': loss / len(ims), 'acc': (acc_i + acc_t) / 2 / len(image) / len(ims)}, prog_bar=True)
            self.logger.log_metrics({"loss": float(loss) / len(ims), "acc": float(acc_i + acc_t) / 2 / len(image) / len(ims)})

        if isinstance(optimizer, list):
            optimizer = optimizer[0]
        optimizer.zero_grad()

        for param in self.image_encoder.parameters():
            param.requires_grad_(False)

        # text loss
        for j, mb in enumerate(text_mbs):
            text_tmp = copy.deepcopy(txt)
            text_tmp[self.global_rank][j*self.minibatch_size:(j+1)*self.minibatch_size] = F.normalize(self.change_text_emb(self.encode_text(mb)), dim=1)
            image_logits_notemp = torch.cat(ims) @ torch.cat(text_tmp).t()
            image_logits = image_logits_notemp * self.model.logit_scale.exp()
            loss = (F.cross_entropy(image_logits, ground_truth) + F.cross_entropy(image_logits.t(), ground_truth))/2
            loss += (F.kl_div(image_logits_notemp * self.sink_temp, img_target) + F.kl_div(image_logits_notemp.t() * self.sink_temp, txt_target)) / 2 * self.kl_coeff
            self.manual_backward(loss)

        optimizer.step()
        lr_scheduler = self.lr_schedulers()
        lr_scheduler.step()
        self.model.logit_scale.data.clamp_(-np.log(100), np.log(100))
        self.sink_temp.data.clamp_(-np.log(100), np.log(100))
        self.update_teacher()

    def encode_text(self, inputs, teacher=False):
        if self.avg_word_embs:
            sequence_output = self.teacher.transformer(**inputs)[0] if teacher else self.model.transformer(**inputs)[0]

            embeddings = torch.sum(
                sequence_output * inputs["attention_mask"].unsqueeze(-1), dim=1
            ) / torch.clamp(torch.sum(inputs["attention_mask"], dim=1, keepdims=True), min=1e-9)
            return embeddings
        else:
            return self.teacher.transformer(**inputs)[1] if teacher else self.model.transformer(**inputs)[1]

    def compute_similarities(self, I_emb, T_emb):
        sim_ii, sim_tt = I_emb @ I_emb.t(), T_emb @ T_emb.t()
        sim_it, sim_ti = I_emb @ T_emb.t(), T_emb @ I_emb.t()
        return sim_ii, sim_tt, sim_it, sim_ti

    def update_teacher(self):
        for teacher, student in zip(self.teacher.parameters(), self.model.parameters()):
            teacher.data.copy_(self.ema(teacher.data, student.data))

    def ema(self, s, t):
        return s * (1 - 0.999) + t * 0.999

    def forward(self, images, text):
        #logits = self.image_encoder.encode_image(images) @ F.normalize(self.change_text_emb(self.encode_text(text)), dim=1).t() * self.model.logit_scale.exp()
        logits = F.normalize(self.image_encoder.encode_image(images), dim=1) @ F.normalize(self.change_text_emb(self.encode_text(text)), dim=1).t() * self.model.logit_scale.exp()
        return logits, logits.t()

    # Sourced from: https://github.com/facebookresearch/swav/blob/5e073db0cc69dea22aa75e92bfdd75011e888f28/main_swav.py#L354
    def sinkhorn(self, out):
        Q = torch.exp(out / 0.05).t()  # Q is K-by-B for consistency with notations from our paper
        B = Q.shape[1]  # number of samples to assign
        K = Q.shape[0]  # how many prototypes

        # make the matrix sums to 1
        sum_Q = torch.sum(Q)
        Q /= sum_Q

        for it in range(3):
            # normalize each row: total weight per prototype must be 1/K
            sum_of_rows = torch.sum(Q, dim=1, keepdim=True)
            Q /= sum_of_rows
            Q /= K

            # normalize each column: total weight per sample must be 1/B
            Q /= torch.sum(Q, dim=0, keepdim=True)
            Q /= B

        Q *= B  # the colomns must sum to 1 so that Q is an assignment
        return Q.t()

    def configure_optimizers(self):
        lr = self.learning_rate

        optimizer = torch.optim.SGD(
            self.parameters(),
            lr=lr,
            momentum=0.9
        )

        # TODO Watch: https://github.com/openai/CLIP/issues/107
        lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
            optimizer,
            T_0=2000
        )

        return {'optimizer': optimizer, 'lr_scheduler': lr_scheduler}



import setuptools
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="soco_clip",
    packages = find_packages(),
    package_data={'soco_clip': ['soco_clip/models/configs/*.yaml','*.gz']},
    include_package_data=True,
    version="0.0.10",
    author="kyusonglee",
    description="OpenAI CLIP wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.soco.ai",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
    ],
    install_requires = [
        "pytorch_lightning",
        "transformers",
        "torch",
        "torchvision",
        "ftfy",
        "opencv-python"
    ]
)
